# File: penCQR.R
# penalized weighted CQR
# Author: Jelena Bradic and Weiwei Wang


.libPaths("~/Rlibs")
library(quadprog)
source("util.R");
source("pcqr.R");

# type: how to select lambda
# if using CV, beta0 should be the default value ( i.e. LAD)
penCQR <- function(y,x,tau,weight,lambda,beta0,type=c("test","cv","sic"),nfold=5,s0=0,s1=2,newy,newx)
{
  call <- match.call();
  
  n <- nrow(x);
  p <- ncol(x);
  K <- length(tau);
  

  if(missing(beta0)){beta0 <- pcqr(y,x,0.5)$betas}
  r <- y-x %*% beta0;

  if(missing(weight))
  {
    if(K==1){weight=1;}else{
      btau <- quantile(r, prob=tau);
      ftau <- rep(0,K);
      for(jj in 1:K){ ftau[jj] <- density(r,from=btau[jj],to=btau[jj],n=1)$y;}

      tmp2 <- matrix(tau,nrow=K, ncol=K);
      tmp1 <- t(tmp2);
      tmp <- pmin(tmp1,tmp2)-tmp1*tmp2;
      tmpf <- matrix(ftau,ncol=1);
      m=tmp / (tmpf %*% t(tmpf));   

 
  	  #T=solve(m,diag(K));
  	  #w1=T%*%ftau;

      w1 <- solve.QP(Dmat=m,dvec=rep(0,K),Amat=cbind(matrix(1,ncol=1,nrow=K), diag(K)),bvec=c(1,rep(0,K)),meq=1)$solution;
      w2=w1/ftau;
      weight <- w1/sum(w1);
      
    }
  }



  # choose lambda
  if(missing(lambda))
  {
    vtmp <- c(0,1);
    while((s1-s0>1e-4) & (max(vtmp)-min(vtmp)>1e-4))
      {
        tmp <- seq(s0,s1,(s1-s0)/4);
        print(tmp)
        if(type=="cv"){
          vtmp <- cv.pwcqr(y,x,tau,weight,nfold,tmp)$cv;
        }else if(type=="sic"){
          vtmp <- sic.pwcqr(y,x,tau,weight,beta0,tmp)$sic;
        } else if (type=="test"){
          vtmp <- test.pwcqr(y,x,tau,weight,beta0,tmp,newy,newx)$mse;
          vtmp=as.vector(na.omit(vtmp));
        }
        pos <- which.min(vtmp);
        if(pos==1){s1 <- tmp[2];}
        else if(pos==5){s0 <- tmp[4];}
        else {s0=tmp[pos-1];s1 <- tmp[pos+1];}
      }
    lambda <- s0;
  }
  print(lambda);

  # estimation
  tmp <- pwcqr(y,x,tau,weight,n*scad(lambda, beta0));
  c(tmp,list(lambda=lambda));
}


check.ftn <- function(y,x,beta,btau,tau,weight)
{
  K <- length(tau);
  n <- length(y);
  r <- y-x%*%beta;
  A <- matrix(r, ncol=K,nrow=n) - matrix(btau,ncol=K,nrow=n,byrow=TRUE);
  mean(A%*%(weight*tau) - (A*(A<0))%*%weight);
}

cv.pwcqr <- function(y,x,tau,weight,nfold=5,s)
  {
    K <- nfold;
    n <- length(y);
    
    all.folds <-  split(sample(1:n), rep(1:K, length = n));
    residmat <- matrix(0, length(s), K)
    for (i in seq(K)) {
      omit <- all.folds[[i]]
      beta0 <- pwcqr(y[-omit],x[-omit,],tau=0.5)$beta;
      for (j in 1:length(s)){
        tmp <- pwcqr(y[-omit],x[-omit,],tau=tau, weight=weight, penalty=n*scad(s[j],beta0))
        beta.fit <- tmp$betas;
        btau.fit <- tmp$btau;
        residmat[j, i] <- check.ftn(y[omit],x[omit,],beta.fit,btau.fit,tau,weight);
      }
    }
    cv <- apply(residmat, 1, mean)
    #print(cv)
    cv.error <- sqrt(apply(residmat, 1, var)/K)
    object <- list(s = s, cv = cv, cv.error = cv.error)
  }

sic.pwcqr <- function(y,x,tau,weight,beta0,s)
{
  K <- length(tau);
  n <- length(y);
  sic <- rep(0,length(s));
  df <- sic;
  for(i in 1:length(s)){
    fit <- pwcqr(y,x,tau,weight,n*scad(s[i],beta0));
    
    r <- y-x%*%fit$beta;
    elbowset <- matrix(r,nrow=n,ncol=K) - matrix(fit$btau,nrow=n,ncol=K,byrow=T);
    df[i] <- sum(abs(elbowset)<1e-6);    
    sic[i] <- log(check.ftn(y,x,fit$beta,fit$btau,tau,weight)/K)+log(n*K)/2/n/K*df[i];
  }
  list(s=s,sic=sic,df=df);
}


test.pwcqr <- function(y,x,tau,weight,beta0,s,newy,newx)
{
  K <- length(tau);
  n <- length(y);
  mse <- rep(0,length(s));
  for(i in 1:length(s)){
     fit <- pwcqr(y,x,tau,weight,n*scad(s[i],beta0));
     mse[i] <- check.ftn(newy,newx,fit$beta,fit$btau,tau,weight);
  }
  list(mse=mse, s=s);
}

#lp form of penalized quantiles
pwcqr <- function(y,x,tau,weight=rep(1,length(tau)),penalty=rep(0,ncol(x)))
{
  call <- match.call();
  
  w <- penalty;
    
    require("lpSolve") 
    call <- match.call()
    K<-length(tau)
    p<-dim(x)[2] 
    n<-length(y)  
    cvec<-c(rep(c(0,0,rep(1,n)),K),w,w)/(K*n)
    cvec <- cvec * c(rep(weight,rep(n+2, K)), rep(1,2*p));
    bvec<-c(tau[1]*y,(tau[1]-1)*y) 
    A1<-cbind(rep(1,n),rep(-1,n))
    A2<-rbind(diag(n),diag(n))
    Aleft<-cbind(rbind(tau[1]*A1,(tau[1]-1)*A1),A2)   
    Aright<-rbind(tau[1]*cbind(x,-x),(tau[1]-1)*cbind(x,-x))     
    if(K>=2){
      for(j in 2:K){
        Aleft<-expandmatrix(Aleft,cbind(rbind(tau[j]*A1,(tau[j]-1)*A1),A2) )
        Aright<-rbind(Aright,rbind(tau[j]*cbind(x,-x),(tau[j]-1)*cbind(x,-x)))
        bvec<-c(bvec,c(tau[j]*y,(tau[j]-1)*y) )
      }
    }
    Amat<-cbind(Aleft,Aright)
    lp.out<-lp(direction="min",objective.in=cvec,const.mat=Amat,const.dir=rep(">=",length(bvec)),const.rhs=bvec)  
    Beta<-lp.out$solution
    btau<-tau
    for (j in 1:K){
      a6<-1+(j-1)*(2+n)
      a7<-j*(2+n)
      bbb<-Beta[(a6:a7)]
      btau[j]<-bbb[1]-bbb[2]
    }
    betas1<-rev(rev(Beta)[1:(2*p)]) 
    betas<-betas1[(1:p)]-betas1[((1+p):(2*p))]
    
    object <- list(call = call, tau=tau, penalty=penalty,  betas=betas, btau=btau, status=lp.out$status, weight=weight)
    object
  }

