
# Description: implement LLA of SCAD in weighted lasso, the lambda is
#              automatically selected using 5-fold cross validation
# functions defined in this file:
#
#     scad: first order derivative of SCAD
#    wlasso: weighted lasso
#
# cv.scad.lla: K-fold cross validation to select tuning paramter in scad-lla
# val.scad.lla: choose tuning parameter using a validation (another test) data
# tpscadls: choose lambda using GCV or BIC
# scad.lla: scad lla with specified tuning paramter choosing method

.libPaths("~/Rlibs")
require(quadprog)
source("util.R");
source("wlasso.R");

huber <- function(beta,y,x, knots,weights)
{
  r <- y-x%*%beta;
  n <- length(y);
  obsw <- pmax(matrix(abs(r),ncol=length(knots),nrow=n),
               matrix(knots, ncol=length(knots), nrow=n, byrow=T));

  mean((matrix(r^2, ncol=length(knots),nrow=n)/obsw) %*% weights)
}


# Function: cv.scad.lla
# At the given lambda, compute the CV error
cv.scad.lla.huber <- function(xw,yw,x,y,s,beta0,K=5,knots,weights)
{
  n <- length(y);
  all.folds <- split(sample(1:n), rep(1:K, length = n));
  residmat <- matrix(0, length(s), K);
  
  for (i in 1:K) {
    omit <- all.folds[[i]]
    for (j in 1:length(s)){
      b <- wlasso(x=xw[-omit, ], y=yw[-omit], penalty=n*scad(s[j],beta0));
      residmat[j, i] <- huber(b,y[omit],x[omit,],knots,weights);
    }
  }
  cv <- apply(residmat, 1, mean)
  cv.error <- sqrt(apply(residmat, 1, var)/K)
  object <- list(s = s, cv = cv, cv.error = cv.error)
  object
}

# choose lambda using a validations dataset
val.scad.lla.huber <- function(xw,yw,s,beta0,newy,newx,knots,weights)
{
  n <- length(y);

  resid <- NULL;
  for (j in 1:length(s)){
      b <- wlasso(xw, yw, penalty=n*scad(s[j],beta0));
      resid <- c(resid,  huber(b,newy,newx,knots,weights));
    }
  list(mse=resid,s=s);
}


penHuber <- function(y,x,knots,weights,beta0,lambda,type=c("cv","test"),nfold=5,s0=0,s1=10,newy,newx)
  {
    call <- match.call();
        
    n <- nrow(x);
    p <- ncol(x);

    #if(missing(beta0)){beta0 <- lm(y~x-1)$coef;}
    
    r <- y-x%*%beta0;
    sigma2<-(sd(r))^2;
   
    #if(missing(weights)){weights <- weightL1L2(r,sigma2);}
     if(missing(weights)){weights <- weightHuber(r,knots);}
    
    
    print("weights calculated")
    
    #sigma<-sqrt(sdHuber(r,knots))
    #print(sigma)
    
    tmp <- pmax(matrix(abs(r),ncol=length(knots),nrow=n),
                matrix(knots, ncol=length(knots), nrow=n, byrow=T));
    obs.w <- 1/tmp %*% weights;
    
    require(Matrix)
    no=norm(obs.w);
    ob=obs.w/no;
    #print(sqrt(obs.w))
    #print(sqrt(norm(obs.w))*(1+(ob-1)/2-(ob)^2/8+ob^3/16-5*ob^4/128))
    
    yw <- y*sqrt(norm(obs.w))*(1+(ob-1)/2-(ob)^2/8+ob^3/16-5*ob^4/128);
    xw <- x*sqrt(norm(obs.w))*as.vector(1+(ob-1)/2-(ob)^2/8+ob^3/16-5*ob^4/128);
    
    

    if(missing(lambda))
    {
      vtmp <- c(0,1);
      while((s1-s0>1e-3) & (max(vtmp)-min(vtmp)>1e-3))
        {
          tmp <- seq(s0,s1,(s1-s0)/4);
          print(tmp)
          if(type=="cv"){
            vtmp <- cv.scad.lla.huber(xw,yw,x,y,tmp,beta0,nfold,knots,weights)$cv;
          } else if(type=="test"){
            vtmp <- val.scad.lla.huber(xw,yw,tmp,beta0,newy,newx,knots,weights)$mse;
            vtmp=as.vector(na.omit(vtmp));
          }

#          print(rbind(tmp,vtmp));
          pos <- which.min(vtmp);
          if(pos==1){s1 <- tmp[2];}
          else if(pos==5){s0 <- tmp[4];}
          else {s0=tmp[pos-1];s1 <- tmp[pos+1];}
        }
      lambda <- s0;
    }
print(lambda)

    # estimation using LLA of SCAD penaty with beta0 as initial estimator=Lasso
    penalty <- n*scad(lambda, beta0);
    beta <- wlasso(xw,yw,penalty);
    
    list(beta=beta,lambda=lambda,weights=weights);
  }

weightHuber<- function(x,knots)
{
  n <- length(x);
  K <- length(knots);
  

  tmp1 <- matrix(abs(x),ncol=K,nrow=n);
  tmp2 <- matrix(knots,ncol=K,nrow=n,byrow=T)
  A <- tmp1<=tmp2;
  B=1-A;
  denom=apply(A,2,mean);

  tmp5=(t(A)%*%(x*x))/n;
  tmp6=t(B)%*%(tmp1*A)/n;    
  tmp7=matrix(tmp5,ncol=K,nrow=K)+matrix(knots,ncol=K,nrow=K)*tmp6+matrix(knots,ncol=1)%*%((1-denom)*knots);
  tmp7[lower.tri(tmp7,F)] <- tmp7[upper.tri(tmp7,F)];
  psimat=tmp7;
  
  #T=solve(psimat,diag(K));
  #w1=T%*%denom;
  #w1=w1/sum(w1);
  
  #sigma=(t(denom)%*%T%*%denom)^(-1);
  
  #w1 <- LowRankQP(Vmat=psimat,dvec=rep(0,K),Amat=diag(K),bvec=rep(1,K),method="LU",u=rep(50,K))$alpha;
   w1 <- solve.QP(Dmat=psimat,dvec=rep(0,K),Amat=cbind(denom, diag(K)),bvec=c(1,rep(0,K)),meq=1)$solution;
  weight <- w1/sum(w1);
  
}

sdHuber<-function(x,knots)
{
  n <- length(x);
  K <- length(knots);
  

  tmp1 <- matrix(abs(x),ncol=K,nrow=n);
  tmp2 <- matrix(knots,ncol=K,nrow=n,byrow=T)
  A <- tmp1<=tmp2;
  B=1-A;
  denom=apply(A,2,mean);

  tmp5=(t(A)%*%(x*x))/n;
  tmp6=t(B)%*%(tmp1*A)/n;    
  tmp7=matrix(tmp5,ncol=K,nrow=K)+matrix(knots,ncol=K,nrow=K)*tmp6+matrix(knots,ncol=1)%*%((1-denom)*knots);
  tmp7[lower.tri(tmp7,F)] <- tmp7[upper.tri(tmp7,F)];
  psimat=tmp7;
  
  T=solve(psimat,diag(K));
  #w1=T%*%denom;
  #w1=w1/sum(w1);
  
  sigma=(t(denom)%*%T%*%denom)^(-1);
  
  #w1 <- LowRankQP(Vmat=psimat,dvec=rep(0,K),Amat=diag(K),bvec=rep(1,K),method="LU",u=rep(50,K))$alpha;
   #w1 <- solve.QP(Dmat=psimat,dvec=rep(0,K),Amat=cbind(denom, diag(K)),bvec=c(1,rep(0,K)),meq=1)$solution;
  #weight <- w1/sum(w1);
  
}

cv.scad.lla.huber1 <- function(xw,yw,x,y,s,beta0,K=5,knots,weights)
{
  n <- length(y);
  all.folds <- split(sample(1:n), rep(1:K, length = n));
  residmat <- matrix(0, length(s), K);
  
  for (i in 1:K) {
    omit <- all.folds[[i]]
    for (j in 1:length(s)){
      b <- wlasso(x=xw[-omit, ], y=yw[-omit], penalty=n*scad(s[j],beta0));
      residmat[j, i] <- huber(b,y[omit],x[omit,],knots,weights);
    }
  }
  cv <- apply(residmat, 1, mean)
  cv.error <- sqrt(apply(residmat, 1, var)/K)
  object <- list(s = s, cv = cv, cv.error = cv.error)
  object
}

