# File:scad-lla.R
# Description: implement LLA of SCAD in weighted lasso, the lambda is
#              automatically selected using 5-fold cross validation
# functions defined in this file:
#
#     scad: first order derivative of SCAD
#    wlasso: weighted lasso
# cv.scad.lla: K-fold cross validation to select tuning paramter in scad-lla
# val.scad.lla: choose tuning parameter using a validation (another test) data
# tpscadls: choose lambda using GCV or BIC
# scad.lla: scad lla with specified tuning paramter choosing method


source("util.R");
source("wlasso.R");



# Function: cv.scad.lla
# At the given lambda, compute the CV error
cv.scad.lla <- function(x,y,s,beta0,K=5)
{
  n <- length(y);
  all.folds <- split(sample(1:n), rep(1:K, length = n));
  residmat <- matrix(0, length(s), K);
  
  for (i in 1:K) {
    omit <- all.folds[[i]]
#    beta0 <- lm(y[-omit]~x[-omit,]-1)$coef;
    
    for (j in 1:length(s)){
      b <- wlasso(x=x[-omit, ], y=y[-omit], penalty=n*scad(s[j],beta0));
      residmat[j, i] <- mean((y[omit]-x[omit,]%*%b)^2);
    }
  }
  cv <- apply(residmat, 1, mean)
  cv.error <- sqrt(apply(residmat, 1, var)/K)
  object <- list(s = s, cv = cv, cv.error = cv.error)
  object
}

# choose lambda using a validations dataset
val12.scad.lla <- function(x,y,s,beta0,newy,newx)
{
  n <- length(y);
  resid <- NULL;
  for (j in 1:length(s)){
      b <- wlasso(x, y, penalty=n*scad(s[j],beta0));
      resid <- c(resid, mean((newy-newx%*%b)^2));
    }
  list(mse=resid,s=s);
}

# choose lambda using GCV or BIC
# Lv JinChi's Matlab code
tpscadls <- function(x,y,s,beta0)
  {
    n <- nrow(x);
    p <- ncol(x);

    bic <- gcv <- rep(0,length(s));
    for(k in 1:length(s)){
      lambda <- s[k];
      beta=wlasso(x,y,penalty=n*scad(lambda, beta0));
      mse=mean((y-x%*%beta)^2);

      set1 <- which(abs(beta)>1e-6);
      if(length(set1)>0){
        x0=as.matrix(x[,set1]);
        abeta=abs(beta[set1]);
        pbeta <- scad(lambda,beta[set1]);
        pt=sum(diag((x0%*%solve(t(x0)%*%x0+n*diag(pbeta/abeta,ncol=length(set1)))%*%t(x0))));
      } else{
        pt=p-1;
      }
      gcv[k]=mse/(1-pt/n)/(1-pt/n);
      bic[k] = log(mse) + pt * (log(n)/n);
    }

    list(gcv=gcv, bic=bic);
  }



scad12.lla <- function(y,x,beta0,lambda,type=c("bic","gcv","cv","test"),nfold=5,s0=0,s1=10,newy,newx)
  {
    call <- match.call();
        
    n <- nrow(x);
    p <- ncol(x);
    
     r=y-x%*%beta0;
    knots=quantile(abs(r),c(0.1,1));
    weights <- weightHuber(r,knots);
    #making a new x and y with changed weightes
    for (i in 1:100)
    {
    	x[i,]<-(weights[1]/abs(r[i])+weights[2])*x[i,]
    	#newx[i,]<-(weights[1]/abs(r[i])+weights[2])*newx[i,]
    	}
    #componentwise multiplication
    y<-(weights[1]/abs(r)+weights[2])*y;
    #newy<-(weights[1]/abs(r)+weights[2])*newy;
    #normalizing data
    #x <- scale(x, TRUE, FALSE);
    #y <- drop(y - mean(y));
 	#newx <- scale(newx, TRUE, FALSE);
    #newy <- drop(newy - mean(newy));
    

    
    if(missing(lambda))
    {
      vtmp <- c(0,1);
      while((s1-s0>1e-2) & (max(vtmp)-min(vtmp)>1e-2))
        {
          tmp <- seq(s0,s1,(s1-s0)/4);
          print(tmp)
          if(type=="cv"){
            vtmp <- cv.scad.lla(x,y,tmp,beta0,nfold)$cv;
          }else if(type=="bic"){
            tp <- tpscadls(x,y,tmp,beta0);
            vtmp <- tp$bic;
          } else if (type=="gcv"){
            tp <- tpscadls(x,y,tmp,beta0);
            vtmp <- tp$gcv;
          } else if(type=="test"){
            vtmp <- val12.scad.lla(x,y,tmp,beta0,newy,newx)$mse;
            vtmp=as.vector(na.omit(vtmp))
          }

#          print(rbind(tmp,vtmp));
          pos <- which.min(vtmp);
          if(pos==1){s1 <- tmp[2];}
          else if(pos==5){s0 <- tmp[4];}
          else {s0=tmp[pos-1];s1 <- tmp[pos+1];}

         
        }
      lambda <- s0;
      print(lambda)
    }
    
    # estimation
    penalty <- n*scad(lambda, beta0);
    #print(penalty)
    beta <- wlasso(x,y,penalty);
    list(beta=beta,lambda=lambda,weights=weights);
  }
