# File: wlasso.R
#
# Description: weighted lasso, compute:
#
#              argmin sum_{i=1}^n (y_i-x_i * beta)^2 + n * sum_{j=1}^p lambda_j * |beta_j|
#
#           lambda_j are given, 
#
# Arguments: x -- n by p matrix
#            y -- n by 1 vector
#            penalty -- p by 1 vector, penalty on each beta (i.e. lambda_j)

#
# Assume x and y are centered and x are normalized
# Author: Jelena Bradic and Weiwei Wang

require(lars)

wlassoonestep <- function(x,y,beta0,penalty=rep(1,ncol(x)))
{
  n <- nrow(x);
  p <- ncol(x);
  
  # find out which penalties are exactly zero;
  setU <- which(penalty<1e-6);
  setV <- setdiff(1:p, setU);


  if (length(setV)==0) # set V is empty, all penalty equal to zero
  {
    fit <- lm(y~x-1);
    beta <- fit$coef;
      }
  else
  {
    if (length(setU)==0)
    {
      x1 <- scale(x,FALSE,penalty);
      fit <- lars(x1,y,normalize=F,use.Gram=F,intercept=F,type="lasso",trace=F,max.steps=1);
      beta <- predict.lars(fit,s=1,type="coefficients",mode=c("lambda"))$coef;
      beta <- beta/penalty;
    }
    else
    {
      Xu <- x[,setU];
      Hu <- Xu %*% solve(t(Xu) %*% Xu) %*% t(Xu);
      y2 <- y-Hu%*%y;
      Xv <- scale(x[,setV], FALSE, penalty[setV]);
      Xv2 <- Xv-Hu%*%Xv;
      fit <- lars(Xv2,y2,normalize=F,intercept=F,type="lasso",trace=F,use.Gram=F,max.steps=1);
      betaV <- predict.lars(fit,s=1,type="coefficients",mode=c("lambda"))$coef;
      betaU <- solve(t(Xu)%*%Xu)%*%t(Xu)%*%(y-Xv%*%betaV);
      beta <- rep(0,p);
      beta[setU] <- betaU;
      beta[setV] <- betaV/penalty[setV];
    }
  }  

#  list(beta=beta,fit=fit);
  beta
}
